from django.test import TestCase

from djangooidc import *  # NOQA
from djangooidc.backends import *  # NOQA
from djangooidc.oidc import *  # NOQA
from djangooidc.urls import *  # NOQA
from djangooidc.views import *  # NOQA


class SmokeImportTest(TestCase):

    def test_smoke_import(self):
        # pass if imports on module level are fine
        return



# coding: utf-8
import logging

from django.conf import settings
from django.http import HttpResponseRedirect
from oic import oic, rndstr
from oic.exception import MissingAttribute
from oic.oauth2 import ErrorResponse, MissingEndpoint, ResponseError
from oic.oic import (AuthorizationRequest, AuthorizationResponse,
                     ProviderConfigurationResponse, RegistrationResponse)
from oic.utils.authn.client import CLIENT_AUTHN_METHOD

__author__ = 'roland'

logger = logging.getLogger(__name__)

default_ssl_check = getattr(settings, 'OIDC_VERIFY_SSL', True)


class OIDCError(Exception):
    pass


class Client(oic.Client):

    def __init__(self, client_id=None, ca_certs=None,
                 client_prefs=None, client_authn_method=None, keyjar=None,
                 verify_ssl=True, behaviour=None):
        oic.Client.__init__(self, client_id, ca_certs, client_prefs,
                            client_authn_method, keyjar, verify_ssl)
        if behaviour:
            self.behaviour = behaviour
        else:
            self.behaviour = {}

    def create_authn_request(self, session, *,
                             acr_value=None, extra_args=None):
        session["state"] = rndstr()
        session["nonce"] = rndstr()
        request_args = {
            "response_type": self.behaviour["response_type"],
            "scope": self.behaviour["scope"],
            "state": session["state"],
            "nonce": session["nonce"],
            "redirect_uri": self.registration_response["redirect_uris"][0]
        }

        if acr_value is None:
            acr_value = self.behaviour.get('acr_value')

        if acr_value is not None:
            request_args["acr_values"] = acr_value

        if extra_args is not None:
            request_args.update(extra_args)
        cis = self.construct_AuthorizationRequest(request_args=request_args)
        logger.debug("request: %s" % cis)

        url, body, ht_args, cis = self.uri_and_body(AuthorizationRequest, cis,
                                                    method="GET",
                                                    request_args=request_args,)
        logger.debug("body: %s" % body)
        logger.info("URL: %s" % url)
        logger.debug("ht_args: %s" % ht_args)

        resp = HttpResponseRedirect(str(url))
        if ht_args:
            for key, value in ht_args.items():
                resp[key] = value
        logger.debug("resp_headers: %s" % ht_args)
        return resp

    def callback(self, response, session):
        """
        Should be called when an AuthN response has been received from the OP.

        :param response: The URL returned by the OP
        :return:
        """
        try:
            authresp = self.parse_response(AuthorizationResponse, response,
                                           sformat="dict", keyjar=self.keyjar)
        except ResponseError as e:
            return OIDCError(u"Response error: {}".format(e))

        if isinstance(authresp, ErrorResponse):
            if authresp["error"] == "login_required":
                return self.create_authn_request(session)
            else:
                return OIDCError("Access denied")

        if session["state"] != authresp["state"]:
            return OIDCError("Received state not the same as expected.")

        try:
            if authresp["id_token"] != session["nonce"]:
                return OIDCError("Received nonce not the same as expected.")
            self.id_token[authresp["state"]] = authresp["id_token"]
        except KeyError:
            pass

        if self.behaviour["response_type"] == "code":
            # get the access token
            try:
                args = {
                    "code": authresp["code"],
                    "redirect_uri": self.registration_response["redirect_uris"][0],
                    "client_id": self.client_id,
                    "client_secret": self.client_secret
                }

                atresp = self.do_access_token_request(
                    scope="openid", state=authresp["state"], request_args=args,
                    authn_method=self.registration_response["token_endpoint_auth_method"])
            except Exception as err:
                logger.error("%s" % err)
                raise

            if isinstance(atresp, ErrorResponse):
                raise OIDCError("Invalid response %s." % atresp["error"])
            session['id_token'] = atresp['id_token']._dict
            if session['id_token']:
                session['id_token_raw'] = getattr(self, 'id_token_raw', None)
            session['access_token'] = atresp['access_token']
            try:
                session['refresh_token'] = atresp['refresh_token']
            except:
                pass

        try:
            inforesp = self.do_user_info_request(
                state=authresp["state"], method="GET")

            if isinstance(inforesp, ErrorResponse):
                raise OIDCError("Invalid response %s." % inforesp["error"])

            userinfo = inforesp.to_dict()

            logger.debug("UserInfo: %s" % inforesp)
        except MissingEndpoint as e:
            logging.warning("Wrong OIDC provider implementation or configuration: {}; using token as userinfo source".format(
                e
            ))
            userinfo = session.get('id_token', {})

        return userinfo

    def store_response(self, resp, info):
        # makes raw ID token available for internal means
        try:
            import json
            from oic.oic.message import AccessTokenResponse
            if isinstance(resp, AccessTokenResponse):
                info = json.loads(info)
                self.id_token_raw = info['id_token']
        except Exception as e:
            # fail silently if something is wrong
            logger.exception(e)

        super(Client, self).store_response(resp, info)

    def __repr__(self):
        return u"Client {} {} {}".format(
            self.client_id,
            self.client_prefs,
            self.behaviour,
        )


class OIDCClients(object):

    def __init__(self, config):
        """

        :param config: Imported configuration module
        :return:
        """
        self.client = {}
        self.client_cls = Client
        self.config = config

        for key, val in config.OIDC_PROVIDERS.items():
            if key == "":
                continue
            else:
                self.client[key] = self.create_client(**val)

    def create_client(self, userid="", **kwargs):
        """
        Do an instantiation of a client instance

        :param userid: An identifier of the user
        :param: Keyword arguments
            Keys are ["srv_discovery_url", "client_info", "client_registration",
            "provider_info"]
        :return: client instance
        """
        _key_set = set(kwargs.keys())
        args = {}
        for param in ["verify_ssl"]:
            try:
                args[param] = kwargs[param]
            except KeyError:
                pass
            else:
                _key_set.discard(param)

        try:
            verify_ssl = default_ssl_check
        except:
            verify_ssl = True

        client = self.client_cls(client_authn_method=CLIENT_AUTHN_METHOD,
                                 behaviour=kwargs["behaviour"], verify_ssl=verify_ssl, **args)

        # The behaviour parameter is not significant for the election process
        _key_set.discard("behaviour")
        for param in ["allow"]:
            try:
                setattr(client, param, kwargs[param])
            except KeyError:
                pass
            else:
                _key_set.discard(param)

        if _key_set == set(["client_info"]):  # Everything dynamic
            # There has to be a userid
            if not userid:
                raise MissingAttribute("Missing userid specification")

            # Find the service that provides information about the OP
            issuer = client.wf.discovery_query(userid)
            # Gather OP information
            client.provider_config(issuer)
            # register the client
            client.register(
                client.provider_info["registration_endpoint"],
                **kwargs["client_info"]
            )
        elif _key_set == set(["client_info", "srv_discovery_url"]):
            # Ship the webfinger part
            # Gather OP information
            client.provider_config(kwargs["srv_discovery_url"])
            # register the client
            client.register(
                client.provider_info["registration_endpoint"],
                **kwargs["client_info"]
            )
        elif _key_set == set(["provider_info", "client_info"]):
            client.handle_provider_config(
                ProviderConfigurationResponse(**kwargs["provider_info"]),
                kwargs["provider_info"]["issuer"])
            client.register(client.provider_info["registration_endpoint"],
                            **kwargs["client_info"])
        elif _key_set == set(["provider_info", "client_registration"]):
            client.handle_provider_config(
                ProviderConfigurationResponse(**kwargs["provider_info"]),
                kwargs["provider_info"]["issuer"])
            client.store_registration_info(RegistrationResponse(
                **kwargs["client_registration"]))
        elif _key_set == set(["srv_discovery_url", "client_registration"]):
            try:
                client.provider_config(kwargs["srv_discovery_url"])
                client.store_registration_info(
                    RegistrationResponse(**kwargs["client_registration"])
                )
            except Exception as e:
                logger.error(
                    "Provider info discovery failed for %s - assume backend unworkable",
                    kwargs["srv_discovery_url"]
                )
                logger.exception(e)
        else:
            raise Exception("Configuration error ?")

        return client

    def dynamic_client(self, userid):
        try:
            dyn = settings.OIDC_ALLOW_DYNAMIC_OP or False
        except:
            dyn = True
        if not dyn:
            raise KeyError("No dynamic clients allowed")

        client = self.client_cls(client_authn_method=CLIENT_AUTHN_METHOD,
                                 verify_ssl=default_ssl_check)

        issuer = client.wf.discovery_query(userid)
        if issuer in self.client:
            return self.client[issuer]
        else:
            # Gather OP information
            _pcr = client.provider_config(issuer)
            # register the client
            client.register(_pcr["registration_endpoint"], **
                            self.config.OIDC_DYNAMIC_CLIENT_REGISTRATION_DATA)
            try:
                client.behaviour.update(**self.config.OIDC_DEFAULT_BEHAVIOUR)
            except KeyError:
                pass

            self.client[issuer] = client
            return client

    def __getitem__(self, item):
        """
        Given a service or user identifier return a suitable client
        :param item:
        :return:
        """
        try:
            return self.client[item]
        except KeyError:
            return self.dynamic_client(item)

    def keys(self):
        return self.client.keys()



# coding: utf-8

from __future__ import unicode_literals

import datetime

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.backends import ModelBackend


class OpenIdConnectBackend(ModelBackend):
    """
    This backend checks a previously performed OIDC authentication.
    If it is OK and the user already exists in the database, it is returned.
    If it is OK and user does not exist in the database, it is created and returned unless setting
        OIDC_CREATE_UNKNOWN_USER is False.
    In all other cases, None is returned.
    """

    def authenticate(self, **kwargs):
        user = None
        if not kwargs or 'sub' not in kwargs.keys():
            return user

        UserModel = get_user_model()
        username = self.clean_username(kwargs['sub'])
        if 'upn' in kwargs.keys():
            username = kwargs['upn']

        # Some OP may actually choose to withhold some information, so we must
        # test if it is present
        openid_data = {'last_login': datetime.datetime.now()}
        if 'first_name' in kwargs.keys():
            openid_data['first_name'] = kwargs['first_name']
        if 'given_name' in kwargs.keys():
            openid_data['first_name'] = kwargs['given_name']
        if 'christian_name' in kwargs.keys():
            openid_data['first_name'] = kwargs['christian_name']
        if 'family_name' in kwargs.keys():
            openid_data['last_name'] = kwargs['family_name']
        if 'last_name' in kwargs.keys():
            openid_data['last_name'] = kwargs['last_name']
        if 'email' in kwargs.keys():
            openid_data['email'] = kwargs['email']

        # Note that this could be accomplished in one try-except clause, but
        # instead we use get_or_create when creating unknown users since it has
        # built-in safeguards for multiple threads.
        if getattr(settings, 'OIDC_CREATE_UNKNOWN_USER', True):
            args = {UserModel.USERNAME_FIELD: username,
                    'defaults': openid_data, }
            user, created = UserModel.objects.update_or_create(**args)
            if created:
                user = self.configure_user(user, **kwargs)
        else:
            try:
                user = UserModel.objects.get_by_natural_key(username)
            except UserModel.DoesNotExist:
                return None
        return user

    def clean_username(self, username):
        """
        Performs any cleaning on the "username" prior to using it to get or
        create the user object.  Returns the cleaned username.

        By default, returns the username unchanged.
        """
        return username

    def configure_user(self, user):
        """
        Configures a user after creation and returns the updated user.

        By default, returns the user unmodified.
        """
        return user



# coding: utf-8

import logging
from urllib.parse import parse_qs, urlencode

from django import forms
from django.conf import settings
from django.contrib.auth import logout as auth_logout
from django.contrib.auth import authenticate, login
from django.contrib.auth.forms import AuthenticationForm
from django.contrib.auth.views import login as auth_login_view
from django.contrib.auth.views import logout as auth_logout_view
from django.http import Http404, HttpResponseRedirect
from django.shortcuts import redirect, render_to_response, resolve_url
from djangooidc.oidc import OIDCClients, OIDCError

logger = logging.getLogger(__name__)

CLIENTS = OIDCClients(settings)


def template_view_error(request, ctx, **kwargs):
    return render_to_response("djangooidc/error.html", ctx)


view_error_handler = template_view_error


# Step 1: provider choice (form). Also - Step 2: redirect to OP. (Step 3
# is OP business.)


class DynamicProvider(forms.Form):
    hint = forms.CharField(
        required=True, label='OpenID Connect full login', max_length=250)


def openid(request, op_name=None):
    client = None
    request.session["next"] = request.GET[
        "next"] if "next" in request.GET.keys() else "/"
    try:
        dyn = settings.OIDC_ALLOW_DYNAMIC_OP or False
    except:
        dyn = True

    try:
        template_name = settings.OIDC_LOGIN_TEMPLATE
    except AttributeError:
        template_name = 'djangooidc/login.html'

    # Internal login?
    if request.method == 'POST' and "internal_login" in request.POST:
        ilform = AuthenticationForm(request.POST)
        return auth_login_view(request)
    else:
        ilform = AuthenticationForm()

    # Try to find an OP client either from the form or from the op_name URL
    # argument
    if request.method == 'GET' and op_name is not None:
        try:
            client = CLIENTS[op_name]
        except KeyError as e:
            logger.info(str(e))
            raise Http404("OIDC client not found")
        request.session["op"] = op_name

    if request.method == 'POST' and dyn:
        form = DynamicProvider(request.POST)
        if form.is_valid():
            try:
                client = CLIENTS.dynamic_client(form.cleaned_data["hint"])
                request.session["op"] = client.provider_info["issuer"]
            except Exception as e:
                logger.exception("could not create OOID client")
                return view_error_handler(request, {"error": e})
    else:
        form = DynamicProvider()

    # If we were able to determine the OP client, just redirect to it with an
    # authentication request
    if client:
        try:
            return client.create_authn_request(request.session)
        except Exception as e:
            return view_error_handler(request, {"error": e})

    # Otherwise just render the list+form.
    return render_to_response(template_name,
                              {"op_list": [i for i in settings.OIDC_PROVIDERS.keys() if i], 'dynamic': dyn,
                               'form': form, 'ilform': ilform, "next": request.session["next"]})


# Step 4: analyze the token returned by the OP
def authz_cb(request):
    op = request.session.get("op")
    if op is None:
        # client session has been dropped or never existed - just ask him to do
        # it again
        return view_error_handler(request, {
            "error": 'Wrong authentication; Please try again',
            "callback": None
        })
    client = CLIENTS[request.session["op"]]
    query = None

    try:
        query = parse_qs(request.GET.urlencode())
        callback_result = client.callback(query, request.session)
        if isinstance(callback_result, OIDCError):
            raise callback_result
        userinfo = callback_result
        request.session["userinfo"] = userinfo
        user = authenticate(request=request, **userinfo)
        if user:
            login(request, user)
            return redirect(request.session.get("next", "/"))
        else:
            raise Exception('this login is not valid in this application')
    except OIDCError as e:
        return view_error_handler(request, {"error": e, "callback": query})


def logout(request, next_page=None):
    # User is by default NOT redirected to the app - it stays on an OP page after logout.
    # Here we determine if a redirection to the app was asked for and is
    # possible.
    if next_page is None and "next" in request.GET.keys():
        next_page = request.GET['next']
    if next_page is None and "next" in request.session.keys():
        next_page = request.session['next']
    if next_page is None:
        next_page = getattr(settings, 'LOGOUT_REDIRECT_URL', None)

    if "op" not in request.session.keys():
        return auth_logout_view(request, next_page)

    client = None
    try:
        client = CLIENTS[request.session["op"]]
    except KeyError:
        logger.info("OIDC client {} not found".format(request.session["op"]))

    extra_args = {}
    if client is not None and "post_logout_redirect_uris" in client.registration_response.keys() and len(
            client.registration_response["post_logout_redirect_uris"]) > 0:
        if next_page is not None:
            # First attempt a direct redirection from OP to next_page
            next_page_url = resolve_url(next_page)
            urls = [url for url in client.registration_response[
                "post_logout_redirect_uris"] if next_page_url in url]
            if len(urls) > 0:
                extra_args["post_logout_redirect_uri"] = urls[0]
            else:
                # It is not possible to directly redirect from the OP to the page that was asked for.
                # We will try to use the redirection point - if the redirection
                # point URL is registered that is.
                next_page_url = resolve_url('openid_logout_cb')
                urls = [url for url in client.registration_response["post_logout_redirect_uris"] if
                        next_page_url in url]
                if len(urls) > 0:
                    extra_args["post_logout_redirect_uri"] = urls[0]
                else:
                    # Just take the first registered URL as a desperate attempt
                    # to come back to the application
                    extra_args["post_logout_redirect_uri"] = client.registration_response["post_logout_redirect_uris"][
                        0]
    else:
        # No post_logout_redirect_uris registered at the OP - no redirection to
        # the application is possible anyway
        pass

    # Redirect client to the OP logout page
    try:
        # DP HACK: Needed to get logout to actually logout from the OIDC Provider
        # According to ODIC session spec (http://openid.net/specs/openid-connect-session-1_0.html#RPLogout)
        # the user should be directed to the OIDC provider to logout after being
        # logged out here.

        if 'access_token' in request.session and 'state' in request.session:
            # TODO: determine if OIDC was used to login by some better way

            request_args = {
                'id_token_hint': request.session['access_token'],
                'state': request.session['state'],
            }
            # should include the post_logout_redirect_uri
            request_args.update(extra_args)

            url = client.provider_info['end_session_endpoint']
            url += "?" + urlencode(request_args)
            return HttpResponseRedirect(url)

            # Looks like they are implementing back channel logout, without checking for
            # support?
            # http://openid.net/specs/openid-connect-backchannel-1_0.html#Backchannel
            """
            request_args = None
            if 'id_token' in request.session.keys():
                request_args = {'id_token': IdToken(**request.session['id_token'])}
            res = client.do_end_session_request(state=request.session["state"],
                                                extra_args=extra_args, request_args=request_args)
            content_type = res.headers.get("content-type", "text/html") # In case the logout response doesn't set content-type (Seen with Keycloak)
            resp = HttpResponse(content_type=content_type, status=res.status_code, content=res._content)
            for key, val in res.headers.items():
                resp[key] = val
            return resp
            """
    finally:
        # Always remove Django session stuff - even if not logged out from OP.
        # Don't wait for the callback as it may never come.
        auth_logout(request)
        if next_page:
            request.session['next'] = next_page

    return HttpResponseRedirect(next_page)


def logout_cb(request):
    """Simple redirection view: after logout, just redirect to a parameter value inside the session"""
    next = request.session["next"] if "next" in request.session.keys() else "/"
    return redirect(next)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import sys

import djangooidc

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

version = djangooidc.__version__

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

if sys.argv[-1] == 'publish':
    os.system('python setup.py sdist upload')
    print("You probably want to also tag the version now:")
    print("  git tag -a %s -m 'version %s'" % (version, version))
    print("  git push --tags")
    sys.exit()

readme = open('README.rst').read()
history = open('HISTORY.rst').read().replace('.. :changelog:', '')

setup(
    name='django-oidc-tf',
    version=version,
    description="""A Django OpenID Connect (OIDC) authentication backend""",
    long_description=readme + '\n\n' + history,
    author='Thomas Frössman',
    author_email='thomasf@jossystem.se',
    url='https://github.com/thomasf/django-oidc',
    packages=[
        'djangooidc',
    ],
    include_package_data=True,
    install_requires=[
        'django>=1.10',
        'oic>=0.10.0',
    ],
    license="Apache Software License",
    zip_safe=False,
    keywords='repo_name',
    classifiers=[
        'Environment :: Web Environment',
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.5',
    ],
)



# coding: utf-8

from django.conf.urls import url

from . import views

urlpatterns = [
    url(r'^login/?$', views.openid, name='openid'),
    url(r'^openid/(?P<op_name>[^/]+)/?$',
        views.openid, name='openid_with_op_name'),
    url(r'^callback/login/?$', views.authz_cb, name='openid_login_cb'),
    url(r'^logout/?$', views.logout, name='logout'),
    url(r'^callback/logout/?$', views.logout_cb, name='openid_logout_cb'),
]
